"""Scenario compilation."""

from collections.abc import Mapping

from preacher.core.scenario import Scenario
from .error import CompilationError
from .request import compile as compile_request
from .response_description import compile as compile_response_description
from .util import run_on_key


_KEY_LABEL = 'label'
_KEY_REQUEST = 'request'
_KEY_RESPONSE = 'response'


def compile_scenario(obj: Mapping) -> Scenario:
    """
    >>> from unittest.mock import patch, sentinel
    >>> request_patch = patch(
    ...     f'{__name__}.compile_request',
    ...     return_value=sentinel.request,
    ... )
    >>> response_description_patch = patch(
    ...     f'{__name__}.compile_response_description',
    ...     return_value=sentinel.response_description
    ... )

    >>> with request_patch as request_mock, \\
    ...      response_description_patch as response_description_mock:
    ...     scenario = compile_scenario({})
    ...     request_mock.assert_called_once_with({})
    ...     response_description_mock.assert_called_once_with({})
    >>> scenario.request
    sentinel.request
    >>> scenario.response_description
    sentinel.response_description

    >>> compile_scenario({'label': []})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Scenario.label ...: label

    >>> compile_scenario({'request': []})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Scenario.request ...: request

    >>> with request_patch as request_mock:
    ...     request_mock.side_effect=CompilationError(
    ...         message='message',
    ...         path=['foo'],
    ...     )
    ...     compile_scenario({})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: message: request.foo

    >>> with request_patch:
    ...     compile_scenario({'response': 'str'})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Scenario.response...: response

    >>> with request_patch as request_mock, \\
    ...      response_description_patch as response_description_mock:
    ...     response_description_mock.side_effect=CompilationError(
    ...         message='message',
    ...         path=['bar'],
    ...     )
    ...     scenario = compile_scenario({'request': '/path'})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: message: response.bar

    >>> with request_patch as request_mock, \\
    ...      response_description_patch as response_description_mock:
    ...     scenario = compile_scenario({'request': '/path'})
    ...     request_mock.assert_called_once_with('/path')
    >>> scenario.label
    >>> scenario.request
    sentinel.request

    >>> with request_patch as request_mock, \\
    ...      response_description_patch as response_description_mock:
    ...     scenario = compile_scenario({
    ...         'label': 'label',
    ...         'request': {'path': '/path'},
    ...         'response': {'key': 'value'},
    ...     })
    ...     request_mock.assert_called_once_with({'path': '/path'})
    ...     response_description_mock.assert_called_once_with({'key': 'value'})
    >>> scenario.label
    'label'
    >>> scenario.request
    sentinel.request
    >>> scenario.response_description
    sentinel.response_description
    """
    label = obj.get(_KEY_LABEL)
    if label is not None and not isinstance(label, str):
        raise CompilationError(
            message=f'Scenario.{_KEY_LABEL} must be a string',
            path=[_KEY_LABEL],
        )

    request_obj = obj.get(_KEY_REQUEST, {})
    if (
        not isinstance(request_obj, Mapping)
        and not isinstance(request_obj, str)
    ):
        raise CompilationError(
            message=f'Scenario.{_KEY_REQUEST} must be a string or a mapping',
            path=[_KEY_REQUEST],
        )
    request = run_on_key(_KEY_REQUEST, compile_request, request_obj)

    response_obj = obj.get('response', {})
    if not isinstance(response_obj, Mapping):
        raise CompilationError(
            message=f'Scenario.{_KEY_RESPONSE} object must be a mapping',
            path=[_KEY_RESPONSE],
        )
    response_description = run_on_key(
        _KEY_RESPONSE,
        compile_response_description,
        response_obj,
    )

    return Scenario(
        label=label,
        request=request,
        response_description=response_description,
    )



"""Description compilation."""

from collections.abc import Mapping

from preacher.core.description import Description
from .error import CompilationError
from .predicate import compile as compile_predicate
from .extraction import compile as compile_extraction
from .util import run_on_key, map_on_key


_KEY_DESCRIBE = 'describe'
_KEY_IT_SHOULD = 'it_should'


def compile(obj: Mapping) -> Description:
    """
    >>> from unittest.mock import call, patch, sentinel
    >>> extraction_patch = patch(
    ...     f'{__name__}.compile_extraction',
    ...     return_value=sentinel.extraction,
    ... )
    >>> predicate_patch = patch(
    ...     f'{__name__}.compile_predicate',
    ...     return_value=sentinel.predicate,
    ... )

    >>> compile({})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Description.describe ...

    >>> with extraction_patch as extraction_mock, \\
    ...      predicate_patch as predicate_mock:
    ...     description = compile({
    ...         'describe': 'foo',
    ...         'it_should': 'string',
    ...     })
    ...     extraction_mock.assert_called_with('foo')
    ...     predicate_mock.assert_called_once_with('string')
    >>> description.extraction
    sentinel.extraction
    >>> description.predicates
    [sentinel.predicate]

    >>> with extraction_patch as extraction_mock, \\
    ...      predicate_patch as predicate_mock:
    ...     description = compile({
    ...         'describe': 'foo',
    ...         'it_should': {'key': 'value'}
    ...     })
    ...     extraction_mock.assert_called_once_with('foo')
    ...     predicate_mock.assert_called_once_with({'key': 'value'})
    >>> description.extraction
    sentinel.extraction
    >>> description.predicates
    [sentinel.predicate]

    >>> with extraction_patch as extraction_mock, \\
    ...      predicate_patch as predicate_mock:
    ...     description = compile({
    ...         'describe': {'key': 'value'},
    ...         'it_should': [{'key1': 'value1'}, {'key2': 'value2'}]
    ...     })
    ...     extraction_mock.assert_called_once_with({'key': 'value'})
    ...     predicate_mock.assert_has_calls([
    ...         call({'key1': 'value1'}),
    ...         call({'key2': 'value2'}),
    ...     ])
    >>> description.extraction
    sentinel.extraction
    >>> description.predicates
    [sentinel.predicate, sentinel.predicate]
    """
    extraction_obj = obj.get(_KEY_DESCRIBE)
    if (
        not isinstance(extraction_obj, Mapping)
        and not isinstance(extraction_obj, str)
    ):
        raise CompilationError(
            message='Description.describe must be a mapping',
            path=[_KEY_DESCRIBE],
        )
    extraction = run_on_key(_KEY_DESCRIBE, compile_extraction, extraction_obj)

    predicate_objs = obj.get(_KEY_IT_SHOULD, [])
    if not isinstance(predicate_objs, list):
        predicate_objs = [predicate_objs]
    predicates = list(
        map_on_key(_KEY_IT_SHOULD, compile_predicate, predicate_objs)
    )

    return Description(extraction=extraction, predicates=predicates)



"""Request compilation."""

from collections.abc import Mapping
from typing import Union

from preacher.core.request import Request
from .error import CompilationError


_KEY_PATH = 'path'
_KEY_PARAMS = 'params'


def compile(obj: Union[Mapping, str]) -> Request:
    """
    >>> compile({'path': {'key': 'value'}})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Request.path ...: path

    >>> compile({'params': ''})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Request.params ...: params

    >>> request = compile('/path')
    >>> request.path
    '/path'
    >>> request.params
    {}

    >>> request = compile({})
    >>> request.path
    ''
    >>> request.params
    {}

    >>> request = compile({'path': '/path', 'params': {'key': 'value'}})
    >>> request.path
    '/path'
    >>> request.params
    {'key': 'value'}
    """
    if isinstance(obj, str):
        return compile({_KEY_PATH: obj})

    path = obj.get(_KEY_PATH, '')
    if not isinstance(path, str):
        raise CompilationError(
            message=f'Request.{_KEY_PATH} must be a string',
            path=[_KEY_PATH],
        )

    params = obj.get(_KEY_PARAMS, {})
    if not isinstance(params, Mapping):
        raise CompilationError(
            message=f'Request.{_KEY_PARAMS} must be a mapping',
            path=[_KEY_PARAMS],
        )

    return Request(path=path, params=params)



"""Predicate compilation."""

from typing import Any

from preacher.core.predicate import Predicate, of_hamcrest_matcher
from .matcher import compile as compile_matcher


def compile(obj: Any) -> Predicate:
    """
    >>> from unittest.mock import patch, sentinel

    >>> with patch(
    ...     f'{__name__}.compile_matcher',
    ...     return_value=sentinel.matcher
    ... ) as matcher_mock:
    ...     predicate = compile('matcher')
    ...     matcher_mock.assert_called_with('matcher')
    """
    matcher = compile_matcher(obj)
    return of_hamcrest_matcher(matcher)



"""Extraction compilation."""

from collections.abc import Mapping
from typing import Union

from preacher.core.extraction import Extraction, with_jq
from .error import CompilationError


_EXTRACTION_MAP = {
    'jq': with_jq,
}
_EXTRACTION_KEYS = frozenset(_EXTRACTION_MAP.keys())


def compile(obj: Union[Mapping, str]) -> Extraction:
    """
    >>> compile({})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: ... has 0

    >>> compile('.foo')({'foo': 'bar'})
    'bar'

    >>> compile({'jq': '.foo'})({'foo': 'bar'})
    'bar'
    """
    if isinstance(obj, str):
        return compile({'jq': obj})

    keys = _EXTRACTION_KEYS.intersection(obj.keys())
    if len(keys) != 1:
        raise CompilationError(
            f'Extraction must have only 1 valid key, but has {len(keys)}'
        )
    key = next(iter(keys))

    return _EXTRACTION_MAP[key](obj[key])



"""Response description compilations."""

from collections.abc import Mapping
from typing import Any

from preacher.core.description import Description
from preacher.core.response_description import ResponseDescription
from .error import CompilationError
from .description import compile as compile_description
from .predicate import compile as compile_predicate
from .util import map_on_key


_KEY_STATUS_CODE = 'status_code'
_KEY_BODY = 'body'


def compile(obj: Mapping) -> ResponseDescription:
    """
    >>> from unittest.mock import call, patch, sentinel
    >>> predicate_patch = patch(
    ...     f'{__name__}.compile_predicate',
    ...     return_value=sentinel.predicate,
    ... )
    >>> description_patch = patch(
    ...     f'{__name__}.compile_description',
    ...     return_value=sentinel.description,
    ... )

    >>> with predicate_patch as predicate_mock, \\
    ...      description_patch as description_mock:
    ...     response_description = compile({})
    ...     predicate_mock.assert_not_called()
    ...     description_mock.assert_not_called()
    >>> response_description.status_code_predicates
    []
    >>> response_description.body_descriptions
    []

    >>> compile({'body': 'str'})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: ...: body

    >>> compile({'body': ['str']})
    Traceback (most recent call last):
        ...
    preacher.compilation.error.CompilationError: Description ...: body[0]

    >>> with predicate_patch as predicate_mock, \\
    ...      description_patch as description_mock:
    ...     response_description = compile({
    ...         'status_code': 402,
    ...         'body': {'key1': 'value1'}}
    ...     )
    ...     predicate_mock.assert_called_once_with(402)
    ...     description_mock.assert_called_once_with({'key1': 'value1'})
    >>> response_description.body_descriptions
    [sentinel.description]

    >>> with predicate_patch as predicate_mock, \\
    ...      description_patch as description_mock:
    ...     response_description = compile({
    ...         'status_code': [{'be_greater_than': 0}, {'be_less_than': 400}],
    ...         'body': [{'key1': 'value1'}, {'key2': 'value2'}],
    ...     })
    ...     predicate_mock.assert_has_calls([
    ...         call({'be_greater_than': 0}),
    ...         call({'be_less_than': 400}),
    ...     ])
    ...     description_mock.assert_has_calls([
    ...         call({'key1': 'value1'}),
    ...         call({'key2': 'value2'}),
    ...     ])
    >>> response_description.body_descriptions
    [sentinel.description, sentinel.description]
    """
    status_code_predicate_objs = obj.get(_KEY_STATUS_CODE, [])
    if not isinstance(status_code_predicate_objs, list):
        status_code_predicate_objs = [status_code_predicate_objs]
    status_code_predicates = list(map_on_key(
        key=_KEY_STATUS_CODE,
        func=compile_predicate,
        items=status_code_predicate_objs,
    ))

    body_description_objs = obj.get(_KEY_BODY, [])
    if isinstance(body_description_objs, Mapping):
        body_description_objs = [body_description_objs]
    if not isinstance(body_description_objs, list):
        raise CompilationError(
            message='ResponseDescription.body must be a list or a mapping',
            path=[_KEY_BODY],
        )
    body_descriptions = list(map_on_key(
        key=_KEY_BODY,
        func=_compile_description,
        items=body_description_objs,
    ))

    return ResponseDescription(
        status_code_predicates=status_code_predicates,
        body_descriptions=body_descriptions,
    )


def _compile_description(obj: Any) -> Description:
    if not isinstance(obj, Mapping):
        raise CompilationError('Description must be a mapping')
    return compile_description(obj)
